package at.ac.uibk.gitsearch.config;

import co.elastic.clients.elasticsearch.ElasticsearchClient;
import co.elastic.clients.json.jackson.JacksonJsonpMapper;
import co.elastic.clients.transport.ElasticsearchTransport;
import co.elastic.clients.transport.rest_client.RestClientTransport;
import com.fasterxml.jackson.core.JsonGenerator;
import com.fasterxml.jackson.databind.DeserializationContext;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.SerializerProvider;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.time.Instant;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.time.format.DateTimeParseException;
import java.util.ArrayList;
import java.util.Arrays;
import org.apache.http.HttpHost;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.elasticsearch.client.RestClient;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.core.convert.converter.Converter;
import org.springframework.data.convert.ReadingConverter;
import org.springframework.data.convert.WritingConverter;
import org.springframework.data.elasticsearch.core.convert.ElasticsearchCustomConversions;
import org.springframework.util.Assert;

@Configuration
public class ElasticsearchConfiguration {

    @Value("${spring.elasticsearch.rest.uris}")
    private String elasticSearchUrls;

    @Bean
    //    @Override
    public ElasticsearchCustomConversions elasticsearchCustomConversions() {
        return new ElasticsearchCustomConversions(
            Arrays.asList(
                new ZonedDateTimeWritingConverter(),
                new ZonedDateTimeReadingConverter(),
                new InstantWritingConverter(),
                new InstantReadingConverter(),
                new LocalDateWritingConverter(),
                new LocalDateReadingConverter()
            )
        );
    }

    @WritingConverter
    static class ZonedDateTimeWritingConverter implements Converter<ZonedDateTime, String> {

        @Override
        public String convert(ZonedDateTime source) {
            if (source == null) {
                return null;
            }
            return source.toInstant().toString();
        }
    }

    @ReadingConverter
    static class ZonedDateTimeReadingConverter implements Converter<String, ZonedDateTime> {

        @Override
        public ZonedDateTime convert(String source) {
            if (source == null) {
                return null;
            }
            return Instant.parse(source).atZone(ZoneId.systemDefault());
        }
    }

    @WritingConverter
    static class InstantWritingConverter implements Converter<Instant, String> {

        @Override
        public String convert(Instant source) {
            if (source == null) {
                return null;
            }
            return source.toString();
        }
    }

    @ReadingConverter
    static class InstantReadingConverter implements Converter<String, Instant> {

        @Override
        public Instant convert(String source) {
            if (source == null) {
                return null;
            }
            return Instant.parse(source);
        }
    }

    @WritingConverter
    static class LocalDateWritingConverter implements Converter<LocalDate, String> {

        @Override
        public String convert(LocalDate source) {
            if (source == null) {
                return null;
            }
            return source.toString();
        }
    }

    @ReadingConverter
    static class LocalDateReadingConverter implements Converter<String, LocalDate> {

        @Override
        public LocalDate convert(String source) {
            if (source == null) {
                return null;
            }
            return LocalDate.parse(source);
        }
    }

    public static class MoreRobustLocalDateTimeConverter extends com.fasterxml.jackson.databind.JsonDeserializer<LocalDateTime> {

        @SuppressWarnings("unused")
        private final Logger log = LogManager.getLogger(MoreRobustLocalDateTimeConverter.class);

        @Override
        @SuppressWarnings("PMD.PreserveStackTrace")
        public LocalDateTime deserialize(com.fasterxml.jackson.core.JsonParser p, DeserializationContext ctxt) throws IOException {
            try {
                final String text = p.getText();
                return LocalDateTime.parse(text, DateTimeFormatter.ISO_ZONED_DATE_TIME);
            } catch (DateTimeParseException e) {
                try {
                    final String text = p.getText();
                    return LocalDateTime.parse(text);
                } catch (DateTimeParseException e2) {
                    throw new IOException(e2);
                }
            }
        }
    }

    public static class InstantTimeDeserializer extends com.fasterxml.jackson.databind.JsonDeserializer<Instant> {

        @SuppressWarnings("unused")
        private final Logger log = LogManager.getLogger(InstantTimeDeserializer.class);

        @Override
        @SuppressWarnings("PMD.PreserveStackTrace")
        public Instant deserialize(com.fasterxml.jackson.core.JsonParser p, DeserializationContext ctxt) throws IOException {
            try {
                final String text = p.getText();
                return Instant.parse(text);
            } catch (DateTimeParseException e) {
                try {
                    String text = p.getText();
                    log.warn("Cannot parse  {} as an instant (missing trailing 'Z'?): {}", text, e.getMessage());
                    if (!text.endsWith("Z")) {
                        text = text + "Z";
                    }
                    return Instant.parse(text);
                } catch (DateTimeParseException e2) {
                    throw new IOException(e2);
                }
            }
        }
    }

    public static class InstantTimeConverter extends com.fasterxml.jackson.databind.JsonSerializer<Instant> {

        @SuppressWarnings("unused")
        private final Logger log = LogManager.getLogger(InstantTimeConverter.class);

        @Override
        public void serialize(Instant value, JsonGenerator gen, SerializerProvider serializers) throws IOException {
            String s = value.toString();
            gen.writeString(s);
        }
    }

    @Bean
    ElasticsearchClient getElasticsearchAPIClient() throws MalformedURLException {
        // Create the transport with a Jackson mapper
        final JacksonJsonpMapper mapper = new JacksonJsonpMapper();

        final JavaTimeModule timeModule = new JavaTimeModule();
        timeModule.addDeserializer(LocalDateTime.class, new MoreRobustLocalDateTimeConverter());
        timeModule.addDeserializer(Instant.class, new InstantTimeDeserializer());
        timeModule.addSerializer(Instant.class, new InstantTimeConverter());

        mapper.objectMapper().registerModule(timeModule);
        mapper.objectMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

        ElasticsearchTransport transport = new RestClientTransport(buildClient(), mapper);

        // And create the API client
        return new ElasticsearchClient(transport);
    }

    private RestClient buildClient() throws MalformedURLException {
        Assert.hasText(elasticSearchUrls, "[Assertion Failed] At least one host must be set.");

        ArrayList<HttpHost> httpHosts = new ArrayList<>();
        for (String host : elasticSearchUrls.split(",")) {
            URL hostUrl = new URL(host);
            httpHosts.add(new HttpHost(hostUrl.getHost(), hostUrl.getPort(), hostUrl.getProtocol()));
        }
        return RestClient.builder(httpHosts.toArray(new HttpHost[0])).build();
    }
}
