package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.domain.ChatMessage;
import at.ac.uibk.gitsearch.domain.Review;
import at.ac.uibk.gitsearch.domain.ReviewRating;
import at.ac.uibk.gitsearch.domain.User;
import at.ac.uibk.gitsearch.repository.jpa.ChatMessageRepository;
import at.ac.uibk.gitsearch.repository.jpa.ReviewRepository;
import at.ac.uibk.gitsearch.service.dto.ChatMessageDTO;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Base64;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

@Service
@SuppressWarnings("PMD")
public class ChatMessageService {

    private static final String CREATOR = "Creator";

    private final ChatMessageRepository chatMessageRepository;

    private final ReviewRepository reviewRepository;

    private final UserService userService;

    private final Logger log = LoggerFactory.getLogger(ChatMessageService.class);

    private static final String YOU = "You";

    public ChatMessageService(ChatMessageRepository chatMessageRepository, ReviewRepository reviewRepository, UserService userService) {
        this.chatMessageRepository = chatMessageRepository;
        this.reviewRepository = reviewRepository;
        this.userService = userService;
    }

    public List<ChatMessage> getAllChatMessages() {
        return chatMessageRepository.findAll();
    }

    public Optional<ChatMessage> getChatMessageById(Long id) {
        return chatMessageRepository.findById(id);
    }

    public ChatMessage saveChatMessage(ChatMessageDTO chatMessageDTO) {
        log.info("Attempting to save chat message from DTO: {}", chatMessageDTO);

        ChatMessage chatMessage = new ChatMessage();

        Review review = reviewRepository.getById(chatMessageDTO.getReviewID());

        chatMessage.setReview(review);

        chatMessage.setTimestamp(new Date());
        chatMessage.setMessageText(chatMessageDTO.getMessageText());

        Optional<User> userOptional = userService.getUserWithAuthorities();
        if (!userOptional.isPresent()) {
            log.warn("No logged-in user found while saving chat message.");
            throw new IllegalStateException("No logged-in user found");
        }
        chatMessage.setSender(userOptional.get());

        ChatMessage savedChatMessage = chatMessageRepository.save(chatMessage);
        log.info("Chat message saved with ID: {}", savedChatMessage.getId());

        return savedChatMessage;
    }

    public void deleteChatMessage(Long id) {
        chatMessageRepository.deleteById(id);
    }

    public List<ChatMessage> findByReview(Review review) {
        return chatMessageRepository.findByReview(review);
    }

    public List<ChatMessage> findByReviewOrderByTimestampAsc(Review review) {
        return chatMessageRepository.findByReviewOrderByTimestampAsc(review);
    }

    public boolean hasAccessToReview(User user, Review review) {
        // Check if the User requested the Review
        if (user.getId().equals(review.getRequestedBy())) {
            return true;
        }

        // Check if the User has a ReviewRating (commented) on the Review
        for (ReviewRating reviewRating : review.getComments()) {
            if (user.getId().equals(reviewRating.getUser().getId())) {
                return true;
            }
        }

        // If neither condition is met, return false
        return false;
    }

    public ChatMessageDTO toDTO(ChatMessage chatMessage) {
        ChatMessageDTO chatMessageDTO = new ChatMessageDTO();
        chatMessageDTO.setMessageText(chatMessage.getMessageText());
        chatMessageDTO.setReviewID(chatMessage.getReview().getId());
        chatMessageDTO.setSentByUser(chatMessage.getSender().equals(userService.getUserWithAuthorities().get()));
        if (chatMessage.getSender().equals(userService.getUserWithAuthorities().get())) {
            chatMessageDTO.setSenderName(YOU);
            return chatMessageDTO;
        }
        if (chatMessage.getReview().getRequestedBy().equals(chatMessage.getSender().getId())) {
            chatMessageDTO.setSenderName(CREATOR);
            return chatMessageDTO;
        }

        chatMessageDTO.setSenderName("Reviewer " + generateHash(chatMessage.getSender().getLogin()));

        return chatMessageDTO;
    }

    public String generateHash(String text) {
        try {
            MessageDigest digest = MessageDigest.getInstance("SHA-256");
            byte[] hash = digest.digest(text.getBytes(StandardCharsets.UTF_8));
            String base64hash = Base64.getEncoder().encodeToString(hash);
            return base64hash.substring(0, 4); // Get only the first 4 characters
        } catch (NoSuchAlgorithmException e) {
            log.error("Error while generating hash: {}", e.getMessage());
            return "Error";
        }
    }

    public List<ChatMessageDTO> replaceSenderNamesWithReviewerNames(List<ChatMessageDTO> chatMessageDTOs) {
        // Extract and sort the unique senderNames excluding "You"
        List<String> sortedSenderNames = chatMessageDTOs
            .stream()
            .map(ChatMessageDTO::getSenderName)
            .distinct()
            .filter(senderName -> !senderName.equals(YOU) && !senderName.equals(CREATOR))
            .sorted()
            .collect(Collectors.toList());

        // Map each unique senderName to "Reviewer 1", "Reviewer 2", etc.
        Map<String, String> senderNameToReviewerMap = new ConcurrentHashMap<>();
        int reviewerCount = 1;
        for (String senderName : sortedSenderNames) {
            senderNameToReviewerMap.put(senderName, "Reviewer " + reviewerCount);
            reviewerCount++;
        }

        // Replace senderNames in ChatMessageDTOs
        List<ChatMessageDTO> updatedDTOs = new ArrayList<>();
        for (ChatMessageDTO dto : chatMessageDTOs) {
            ChatMessageDTO updatedDTO = new ChatMessageDTO();
            updatedDTO.setMessageText(dto.getMessageText());
            updatedDTO.setReviewID(dto.getReviewID());
            updatedDTO.setSentByUser(dto.getSentByUser());

            // Replace senderName if it's not "You"
            if (dto.getSenderName().equals(YOU)) {
                updatedDTO.setSenderName(YOU);
            }
            if (dto.getSenderName().equals(CREATOR)) {
                updatedDTO.setSenderName(CREATOR);
            }

            if (!dto.getSenderName().equals(YOU) && !dto.getSenderName().equals(CREATOR)) {
                String newReviewerName = senderNameToReviewerMap.get(dto.getSenderName());
                updatedDTO.setSenderName(newReviewerName);
            }

            updatedDTOs.add(updatedDTO);
        }

        return updatedDTOs;
    }
}
