package at.ac.uibk.gitsearch.service;

import at.ac.uibk.gitsearch.repository.gitlab.GitLabRepository;
import at.ac.uibk.gitsearch.security.jwt.TokenProvider.GitLabAccessInfo;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.JsonMappingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.datatype.jsr310.JavaTimeModule;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import java.io.IOException;
import java.io.InputStream;
import java.time.Duration;
import java.util.ArrayList;
import java.util.List;
import java.util.Optional;
import java.util.UUID;
import java.util.zip.ZipInputStream;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.codeability.sharing.plugins.api.ShoppingBasket;
import org.codeability.sharing.plugins.api.ShoppingBasket.UserInfo;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.gitlab4j.api.GitLabApi;
import org.gitlab4j.api.GitLabApiException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

/**
 * Service for exercise/course search results
 * <p>
 */
@Service
public class ShoppingBasketService {

    public static final Duration BASKET_EXPIRY_INTERVAL = Duration.ofMinutes(60);

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private PluginManagementService pluginManagementService;

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private GitLabRepository gitLabRepository;

    @Autowired
    @SuppressWarnings("PMD.ImmutableField")
    private ZipRepackagingService zipRepackagingService;

    private static final Logger log = LoggerFactory.getLogger(ShoppingBasketService.class);

    static final ObjectMapper objectMapper = new ObjectMapper();

    static {
        objectMapper.registerModule(new JavaTimeModule());
        objectMapper.configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);
    }

    public static class ShoppingBasketRedirectInfoDTO {

        private String redirectURL;

        /**
         * @return the redirectURL
         */
        public String getRedirectURL() {
            return redirectURL;
        }

        /**
         * @param redirectURL the redirectURL to set
         */
        public void setRedirectURL(String redirectURL) {
            this.redirectURL = redirectURL;
        }
    }

    public static class ShoppingBasketInfoDTO {

        private SearchResultDTO[] itemInfos;
        private String plugin;
        private String action;

        public ShoppingBasketInfoDTO(SearchResultDTO[] itemInfos) {
            super();
            this.itemInfos = itemInfos.clone();
        }

        public ShoppingBasketInfoDTO() {
            // JSON
        }

        /**
         * @return the itemInfos
         */
        public SearchResultDTO[] getItemInfos() {
            return itemInfos.clone();
        }

        /**
         * @param itemInfos the itemInfos to set
         */
        public void setItemInfos(SearchResultDTO[] itemInfos) {
            this.itemInfos = itemInfos.clone();
        }

        /**
         * @return the plugin
         */
        public String getPlugin() {
            return plugin;
        }

        /**
         * @param plugin the plugin to set
         */
        public void setPlugin(String plugin) {
            this.plugin = plugin;
        }

        /**
         * @return the action
         */
        public String getAction() {
            return action;
        }

        /**
         * @param action the action to set
         */
        public void setAction(String action) {
            this.action = action;
        }

        public static class ItemInfoDTO {

            String exerciseID;
            String gitLabURL;

            public ItemInfoDTO(String exerciseID, String gitLabURl) {
                super();
                this.exerciseID = exerciseID;
                this.gitLabURL = gitLabURl;
            }

            public ItemInfoDTO() {
                // JSON
            }

            /**
             * @return the exerciseIDs
             */
            public String getExerciseID() {
                return exerciseID;
            }

            /**
             * @param exerciseIDs the exerciseIDs to set
             */
            public void setExerciseID(String exerciseID) {
                this.exerciseID = exerciseID;
            }

            /**
             * @return the gitLabURls
             */
            public String getGitLabURL() {
                return gitLabURL;
            }

            /**
             * @param gitLabURL the gitLabURls to set
             */
            public void setGitLabURL(String gitLabURL) {
                this.gitLabURL = gitLabURL;
            }
        }
    }

    /**
     * holds further info for shopping basket.
     * @author Michael Breu
     *
     */
    public static class ExtendedShoppingBasket {

        private ShoppingBasketInfoDTO shoppingBasket;
        private Optional<GitLabAccessInfo> gitLabAccessInfo;

        /**
         * token valid until (msecs since 1.1.1970)
         */
        private long tokenValidUntil;

        public ExtendedShoppingBasket(
            ShoppingBasketInfoDTO shoppingBasket,
            Optional<GitLabAccessInfo> gitLabAccessInfo,
            long tokenValidUntil
        ) {
            super();
            this.shoppingBasket = shoppingBasket;
            this.gitLabAccessInfo = gitLabAccessInfo;
            this.tokenValidUntil = tokenValidUntil;
        }

        public ExtendedShoppingBasket() {}

        /**
         * @return the shoppingBasket
         */
        public ShoppingBasketInfoDTO getShoppingBasket() {
            return shoppingBasket;
        }

        /**
         * @param shoppingBasket the shoppingBasket to set
         */
        public void setShoppingBasket(ShoppingBasketInfoDTO shoppingBasket) {
            this.shoppingBasket = shoppingBasket;
        }

        /**
         * @return the gitLabAccessInfo
         */
        public Optional<GitLabAccessInfo> getGitLabAccessInfo() {
            return gitLabAccessInfo;
        }

        /**
         * @param gitLabAccessInfo the gitLabAccessInfo to set
         */
        public void setGitLabAccessInfo(Optional<GitLabAccessInfo> gitLabAccessInfo) {
            this.gitLabAccessInfo = gitLabAccessInfo;
        }

        /**
         * @return the tokenValidUntil
         */
        public long getTokenValidUntil() {
            return tokenValidUntil;
        }
    }

    public ShoppingBasket getBasket(String basketToken) {
        if (basketToken == null) {
            log.warn("Basket for null not found");
            return null;
        }
        @Nullable
        final ExtendedShoppingBasket basketInfo = basketCache.getIfPresent(basketToken);
        if (basketInfo == null) {
            log.warn("Basket {} not found", basketToken);
            return null;
        }

        List<SearchResultDTO> result = new ArrayList<>();
        for (SearchResultDTO hit : basketInfo.getShoppingBasket().getItemInfos()) {
            result.add(convertFrom(hit));
        }
        UserInfo userInfo = null;
        final Optional<GitLabAccessInfo> gitLabAccessInfo = basketInfo.getGitLabAccessInfo();
        if (gitLabAccessInfo.isPresent() && gitLabAccessInfo.get().getUser().isPresent()) {
            userInfo = new UserInfo(gitLabAccessInfo.get().getUser().get().getUsername());
        }
        final ShoppingBasket shoppingBasket = new ShoppingBasket(
            userInfo,
            result.toArray(new SearchResultDTO[] {}),
            basketInfo.tokenValidUntil
        );
        log.info("Basket {} is delivered", basketToken);
        return shoppingBasket;
    }

    private SearchResultDTO convertFrom(SearchResultDTO hit) {
        try {
            return objectMapper.readValue(objectMapper.writeValueAsString(hit), SearchResultDTO.class);
        } catch (JsonMappingException e) {
            log.warn("Cannot map hit", e);
            return null;
        } catch (JsonProcessingException e) {
            log.warn("Cannot convert hit", e);
            return null;
        }
    }

    public InputStream getRepositoryZip(String basketToken, int position) throws GitLabApiException, IOException {
        @Nullable
        final ExtendedShoppingBasket basketInfo = basketCache.getIfPresent(basketToken);
        ShoppingBasket basket = getBasket(basketToken);
        final GitLabApi gitLabApi = basketInfo == null
            ? gitLabRepository.getGitLabApi(Optional.empty())
            : gitLabRepository.getGitLabApi(basketInfo.getGitLabAccessInfo());
        return zipRepackagingService.rePackageGitLabProjectZip(
            new ZipInputStream(
                gitLabApi
                    .getRepositoryApi()
                    .getRepositoryArchive(Long.toString(basket.exerciseInfo[position].getProject().getProject_id()), "HEAD", "zip")
            ),
            "from project " + basket.exerciseInfo[position].getProject().getProject_id(),
            Optional.empty(),/* no filters ??? */
            null
        );
    }

    /**
     * this is a cache for shopping baskets. Entries expires after BASKET_EXPIRY_INTERVAL.
     * TODO: Store in Database, to allow multiple application instances.
     */
    protected LoadingCache<String, ExtendedShoppingBasket> basketCache = CacheBuilder
        .newBuilder()
        .maximumSize(10000)
        .expireAfterWrite(BASKET_EXPIRY_INTERVAL)
        .build(
            new CacheLoader<String, ExtendedShoppingBasket>() {
                @Override
                public ExtendedShoppingBasket load(String key) {
                    return null;
                }
            }
        );

    public ShoppingBasketRedirectInfoDTO getRedirectInfo(ShoppingBasketInfoDTO basket, String baseURL) {
        // prepare
        UUID random = UUID.randomUUID();
        long tokenValidUntil = System.currentTimeMillis() + BASKET_EXPIRY_INTERVAL.toMillis();
        basketCache.put(random.toString(), new ExtendedShoppingBasket(basket, gitLabRepository.getGitLabAccessInfo(), tokenValidUntil));

        ShoppingBasketRedirectInfoDTO result = new ShoppingBasketRedirectInfoDTO();

        result.setRedirectURL(
            pluginManagementService.getRedirectURL(basket.getPlugin(), basket.getAction()) +
            "/" +
            random.toString() +
            "?returnURL=" +
            baseURL +
            "&apiBaseURL=" +
            baseURL +
            "/api/pluginIF/v0.1"
        );

        return result;
    }
}
