import { ComponentFixture, TestBed } from '@angular/core/testing';
import { HttpResponse } from '@angular/common/http';
import { HttpClientTestingModule } from '@angular/common/http/testing';
import { FormBuilder } from '@angular/forms';
import { ActivatedRoute } from '@angular/router';
import { RouterTestingModule } from '@angular/router/testing';
import { of, Subject, from } from 'rxjs';

import { ReviewRatingService } from '../service/review-rating.service';
import { IReviewRating, ReviewRating } from '../review-rating.model';
import { IReview } from 'app/entities/review/review.model';
import { ReviewService } from 'app/entities/review/service/review.service';

import { IUser } from 'app/entities/user/user.model';
import { UserService } from 'app/entities/user/user.service';

import { ReviewRatingUpdateComponent } from './review-rating-update.component';

describe('ReviewRating Management Update Component', () => {
  let comp: ReviewRatingUpdateComponent;
  let fixture: ComponentFixture<ReviewRatingUpdateComponent>;
  let activatedRoute: ActivatedRoute;
  let reviewRatingService: ReviewRatingService;
  let reviewService: ReviewService;
  let userService: UserService;

  beforeEach(() => {
    TestBed.configureTestingModule({
      imports: [HttpClientTestingModule, RouterTestingModule.withRoutes([])],
      declarations: [ReviewRatingUpdateComponent],
      providers: [
        FormBuilder,
        {
          provide: ActivatedRoute,
          useValue: {
            params: from([{}]),
          },
        },
      ],
    })
      .overrideTemplate(ReviewRatingUpdateComponent, '')
      .compileComponents();

    fixture = TestBed.createComponent(ReviewRatingUpdateComponent);
    activatedRoute = TestBed.inject(ActivatedRoute);
    reviewRatingService = TestBed.inject(ReviewRatingService);
    reviewService = TestBed.inject(ReviewService);
    userService = TestBed.inject(UserService);

    comp = fixture.componentInstance;
  });

  describe('ngOnInit', () => {
    it('Should call Review query and add missing value', () => {
      const reviewRating: IReviewRating = { id: 456 };
      const review: IReview = { id: 75122 };
      reviewRating.review = review;

      const reviewCollection: IReview[] = [{ id: 42102 }];
      jest.spyOn(reviewService, 'query').mockReturnValue(of(new HttpResponse({ body: reviewCollection })));
      const additionalReviews = [review];
      const expectedCollection: IReview[] = [...additionalReviews, ...reviewCollection];
      jest.spyOn(reviewService, 'addReviewToCollectionIfMissing').mockReturnValue(expectedCollection);

      activatedRoute.data = of({ reviewRating });
      comp.ngOnInit();

      expect(reviewService.query).toHaveBeenCalled();
      expect(reviewService.addReviewToCollectionIfMissing).toHaveBeenCalledWith(reviewCollection, ...additionalReviews);
      expect(comp.reviewsSharedCollection).toEqual(expectedCollection);
    });

    it('Should call User query and add missing value', () => {
      const reviewRating: IReviewRating = { id: 456 };
      const user: IUser = { id: 66647 };
      reviewRating.user = user;

      const userCollection: IUser[] = [{ id: 60979 }];
      jest.spyOn(userService, 'query').mockReturnValue(of(new HttpResponse({ body: userCollection })));
      const additionalUsers = [user];
      const expectedCollection: IUser[] = [...additionalUsers, ...userCollection];
      jest.spyOn(userService, 'addUserToCollectionIfMissing').mockReturnValue(expectedCollection);

      activatedRoute.data = of({ reviewRating });
      comp.ngOnInit();

      expect(userService.query).toHaveBeenCalled();
      expect(userService.addUserToCollectionIfMissing).toHaveBeenCalledWith(userCollection, ...additionalUsers);
      expect(comp.usersSharedCollection).toEqual(expectedCollection);
    });

    it('Should update editForm', () => {
      const reviewRating: IReviewRating = { id: 456 };
      const review: IReview = { id: 79181 };
      reviewRating.review = review;
      const user: IUser = { id: 27448 };
      reviewRating.user = user;

      activatedRoute.data = of({ reviewRating });
      comp.ngOnInit();

      expect(comp.editForm.value).toEqual(expect.objectContaining(reviewRating));
      expect(comp.reviewsSharedCollection).toContain(review);
      expect(comp.usersSharedCollection).toContain(user);
    });
  });

  describe('save', () => {
    it('Should call update service on save for existing entity', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<ReviewRating>>();
      const reviewRating = { id: 123 };
      jest.spyOn(reviewRatingService, 'update').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ reviewRating });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.next(new HttpResponse({ body: reviewRating }));
      saveSubject.complete();

      // THEN
      expect(comp.previousState).toHaveBeenCalled();
      expect(reviewRatingService.update).toHaveBeenCalledWith(reviewRating);
      expect(comp.isSaving).toEqual(false);
    });

    it('Should call create service on save for new entity', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<ReviewRating>>();
      const reviewRating = new ReviewRating();
      jest.spyOn(reviewRatingService, 'create').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ reviewRating });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.next(new HttpResponse({ body: reviewRating }));
      saveSubject.complete();

      // THEN
      expect(reviewRatingService.create).toHaveBeenCalledWith(reviewRating);
      expect(comp.isSaving).toEqual(false);
      expect(comp.previousState).toHaveBeenCalled();
    });

    it('Should set isSaving to false on error', () => {
      // GIVEN
      const saveSubject = new Subject<HttpResponse<ReviewRating>>();
      const reviewRating = { id: 123 };
      jest.spyOn(reviewRatingService, 'update').mockReturnValue(saveSubject);
      jest.spyOn(comp, 'previousState');
      activatedRoute.data = of({ reviewRating });
      comp.ngOnInit();

      // WHEN
      comp.save();
      expect(comp.isSaving).toEqual(true);
      saveSubject.error('This is an error!');

      // THEN
      expect(reviewRatingService.update).toHaveBeenCalledWith(reviewRating);
      expect(comp.isSaving).toEqual(false);
      expect(comp.previousState).not.toHaveBeenCalled();
    });
  });

  describe('Tracking relationships identifiers', () => {
    describe('trackReviewById', () => {
      it('Should return tracked Review primary key', () => {
        const entity = { id: 123 };
        const trackResult = comp.trackReviewById(0, entity);
        expect(trackResult).toEqual(entity.id);
      });
    });

    describe('trackUserById', () => {
      it('Should return tracked User primary key', () => {
        const entity = { id: 123 };
        const trackResult = comp.trackUserById(0, entity);
        expect(trackResult).toEqual(entity.id);
      });
    });
  });
});
